/* $Id: gx_mmio.inc,v 1.4 1998/11/24 18:28:09 marcus Exp $
******************************************************************************

   Cyrix MediaGX I/O functions

   Copyright (C) 1997-1998 Cyrix Corporation
   Copyright (C) 1997-1998 Emmanuel Marty	[core@ggi-project.org]

   Permission is hereby granted, free of charge, to any person obtaining a
   copy of this software and associated documentation files (the "Software"),
   to deal in the Software without restriction, including without limitation
   the rights to use, copy, modify, merge, publish, distribute, sublicense,
   and/or sell copies of the Software, and to permit persons to whom the
   Software is furnished to do so, subject to the following conditions:

   The above copyright notice and this permission notice shall be included in
   all copies or substantial portions of the Software.

   THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
   IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
   FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL
   THE AUTHOR(S) BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
   IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
   CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

******************************************************************************

	This code was derived from the following sources of information :

	Cyrix MediaGX confidential databooks
	Cyrix MediaGX gxvideo.c example sourcecode
	Help and code from Sengan Baring-Gould of Cyrix Corporation

	Many thanks to the Cyrix Corporation for providing me with
	documentation and official support about the GX chipset family.

******************************************************************************
*/

/*
**	write and read external clockchip registers
*/

void CLK_out8 (struct kgi_display *dpy, kgiu8 val, kgiu8 reg) {
	struct	chipset_private *private = KGIM_PRIVATE (dpy, chipset);

	if (!private->companion) {
		mem_vaddr	gxbase = private->gx_control.baseptr;
		kgiu32		lockval, gcfg;

		ASSERT (reg <= 0x0E);

		/*	unlock the display controller registers		*/
		lockval = READ_GX32 (gxbase, DC_UNLOCK);
		WRITE_GX32 (gxbase, DC_UNLOCK, DC_UNLOCK_VALUE);

		/*
		** Before changing the RS lines, the RAMDAC data write buffer
		** must be flushed. This is done by reading from the
		** MC RAMDAC interface
		*/

		gcfg = READ_GX32 (gxbase, MC_RAMDAC_ACC);

		/*	select pll write index register */
		gcfg = READ_GX32 (gxbase, DC_GENERAL_CFG);
		gcfg = (gcfg & ~DC_GCFG_DAC_RS_MASK) |
			( DAC_PLLWRITE << DC_GCFG_DAC_RS_POS );
		WRITE_GX32 (gxbase, DC_GENERAL_CFG, gcfg);

		/*	set pll write index */
		WRITE_GX32 (gxbase, MC_RAMDAC_ACC, reg);

		/*	flush RAMDAC write buffer again			*/
		gcfg = READ_GX32 (gxbase, MC_RAMDAC_ACC);

		/*	select pll data register */
		gcfg = READ_GX32 (gxbase, DC_GENERAL_CFG);
		gcfg = (gcfg & ~DC_GCFG_DAC_RS_MASK) |
			( DAC_PLLDATA << DC_GCFG_DAC_RS_POS );
		WRITE_GX32 (gxbase, DC_GENERAL_CFG, gcfg);

		/*	set pll index data */
		WRITE_GX32 (gxbase, MC_RAMDAC_ACC, val);

		/*	restore status of lock register			*/
		WRITE_GX32 (gxbase, DC_UNLOCK, lockval);
	}
}

kgiu8 CLK_in8 (struct kgi_display *dpy, kgiu8 reg) {
	struct	chipset_private *private = KGIM_PRIVATE (dpy, chipset);

	if (!private->companion) {
		mem_vaddr	gxbase = private->gx_control.baseptr;
		kgiu32		lockval, val, gcfg;

		ASSERT (reg <= 0x0E);

		/*	unlock the display controller registers		*/
		lockval = READ_GX32 (gxbase, DC_UNLOCK);
		WRITE_GX32 (gxbase, DC_UNLOCK, DC_UNLOCK_VALUE);

		/*	flush ramdac write buffer			*/
		gcfg = READ_GX32 (gxbase, MC_RAMDAC_ACC);

		/*	select pll read index register			 */
		gcfg = READ_GX32 (gxbase, DC_GENERAL_CFG);
		gcfg = (gcfg & ~DC_GCFG_DAC_RS_MASK) |
			( DAC_PLLREAD << DC_GCFG_DAC_RS_POS );
		WRITE_GX32 (gxbase, DC_GENERAL_CFG, gcfg);

		/*	set pll read index */
		WRITE_GX32 (gxbase, MC_RAMDAC_ACC, reg);

		/*	flush RAMDAC write buffer again			*/
		gcfg = READ_GX32 (gxbase, MC_RAMDAC_ACC);

		/*	select pll data register */
		gcfg = READ_GX32 (gxbase, DC_GENERAL_CFG);
		gcfg = (gcfg & ~DC_GCFG_DAC_RS_MASK) |
			( DAC_PLLDATA << DC_GCFG_DAC_RS_POS );
		WRITE_GX32 (gxbase, DC_GENERAL_CFG, gcfg);

		/*	read pll index value */
		val = READ_GX32 (gxbase, MC_RAMDAC_ACC);

		/*	restore status of lock register			*/
		WRITE_GX32 (gxbase, DC_UNLOCK, lockval);

		return val;
	}
	else	return 0;
}

/*
**	write and read external RAMDAC registers
*/

void DAC_out8 (struct kgi_display *dpy, kgiu8 val, kgiu8 reg) {
	struct	chipset_private *private = KGIM_PRIVATE (dpy, chipset);

	if (!private->companion) {
		mem_vaddr	gxbase = private->gx_control.baseptr;
		kgiu32		lockval, gcfg;

		ASSERT (!(reg & 0xF8));

		/*	unlock the display controller registers		*/
		lockval = READ_GX32 (gxbase, DC_UNLOCK);
		WRITE_GX32 (gxbase, DC_UNLOCK, DC_UNLOCK_VALUE);

		/*	set the register select lines			*/
		gcfg = READ_GX32 (gxbase, DC_GENERAL_CFG);
		gcfg = (gcfg & ~DC_GCFG_DAC_RS_MASK) |
			( (reg & 0x7) << DC_GCFG_DAC_RS_POS );
		WRITE_GX32 (gxbase, DC_GENERAL_CFG, gcfg);

		/*	write value to ramdac register			*/
		WRITE_GX32 (gxbase, MC_RAMDAC_ACC, val);
	
		/*	restore status of lock register			*/
		WRITE_GX32 (gxbase, DC_UNLOCK, lockval);
	}
}

kgiu8 DAC_in8 (struct kgi_display *dpy, kgiu8 reg) {
	struct	chipset_private *private = KGIM_PRIVATE (dpy, chipset);

	if (!private->companion) {
		mem_vaddr	gxbase = private->gx_control.baseptr;
		kgiu32		lockval, gcfg, val;

		ASSERT (!(reg & 0xF8));

		/*	unlock the display controller registers		*/
		lockval = READ_GX32 (gxbase, DC_UNLOCK);
		WRITE_GX32 (gxbase, DC_UNLOCK, DC_UNLOCK_VALUE);

		/*	set the register select lines			*/
		gcfg = READ_GX32 (gxbase, DC_GENERAL_CFG);
		gcfg = (gcfg & ~DC_GCFG_DAC_RS_MASK) |
		       ( (reg & 0x7) << DC_GCFG_DAC_RS_POS );
		WRITE_GX32 (gxbase, DC_GENERAL_CFG, gcfg);

		/*	read value from ramdac register			*/
		val = READ_GX32 (gxbase, MC_RAMDAC_ACC);
	
		/*	restore status of lock register			*/
		WRITE_GX32 (gxbase, DC_UNLOCK, lockval);

		return ((kgiu8) val);
	}
	else	return 0;
}

/*
**	set line compare (splitline) value
*/

static void gx_set_split (struct kgi_display *dpy, int line) {
	mem_vaddr gxbase = ((struct chipset_private *)
			   KGIM_PRIVATE (dpy,chipset))->gx_control.baseptr;
	kgiu32    lockval;

	/*	unlock the display controller registers		*/
	lockval = READ_GX32 (gxbase, DC_UNLOCK);
	WRITE_GX32 (gxbase, DC_UNLOCK, DC_UNLOCK_VALUE);

	/*	write line compare value			*/
	WRITE_GX32 (gxbase, DC_SS_LINE_COMPARE, line & 0x3FF);

	/*	restore status of lock register			*/
	WRITE_GX32 (gxbase, DC_UNLOCK, lockval);
}

/*
**	set origin
*/

static void gx_set_origin (struct kgi_display *dpy, uint x, uint y) {
	struct	chipset_private *private = KGIM_PRIVATE (dpy, chipset);
	mem_vaddr	gxbase = private->gx_control.baseptr;
	kgiu32    lockval;

	/*	unlock the display controller registers		*/
	lockval = READ_GX32 (gxbase, DC_UNLOCK);
	WRITE_GX32 (gxbase, DC_UNLOCK, DC_UNLOCK_VALUE);

	/*	write origin					*/
	WRITE_GX32 (gxbase, DC_FB_ST_OFFSET, 
			y * private->fblinesize +
			x);	/* << pixshift */

	/*	restore status of lock register			*/
	WRITE_GX32 (gxbase, DC_UNLOCK, lockval);
}

/*
**	pointer functions
*/

static void gx_show_pointer(struct kgi_display *dpy, uint x, uint y) {
	struct	chipset_private *private = KGIM_PRIVATE (dpy, chipset);
	mem_vaddr	gxbase = private->gx_control.baseptr;
	kgiu32	lockval, gcfg;

	/*	unlock the display controller registers		*/
	lockval = READ_GX32 (gxbase, DC_UNLOCK);
	WRITE_GX32 (gxbase, DC_UNLOCK, DC_UNLOCK_VALUE);

	/*	set cursor position				*/
	WRITE_GX32 (gxbase, DC_CURSOR_X, x & 0x7FF);
	WRITE_GX32 (gxbase, DC_CURSOR_Y, y & 0x3FF);

	/*	enable hardware pointer				*/
	gcfg = READ_GX32 (gxbase, DC_GENERAL_CFG);
	gcfg = gcfg | DC_GCFG_CURE;
	WRITE_GX32 (gxbase, DC_GENERAL_CFG, gcfg);

	/*	restore status of lock register			*/
	WRITE_GX32 (gxbase, DC_UNLOCK, lockval);
}

static void gx_hide_pointer(struct kgi_display *dpy) {
	struct	chipset_private *private = KGIM_PRIVATE (dpy, chipset);
	mem_vaddr	gxbase = private->gx_control.baseptr;
	kgiu32	lockval, gcfg;

	/*	unlock the display controller registers		*/
	lockval = READ_GX32 (gxbase, DC_UNLOCK);
	WRITE_GX32 (gxbase, DC_UNLOCK, DC_UNLOCK_VALUE);

	/*	disable hardware pointer			*/
	gcfg = READ_GX32 (gxbase, DC_GENERAL_CFG);
	gcfg = gcfg & ~DC_GCFG_CURE;
	WRITE_GX32 (gxbase, DC_GENERAL_CFG, gcfg);

	/*	restore status of lock register			*/
	WRITE_GX32 (gxbase, DC_UNLOCK, lockval);
}

#define	gx_undo_pointer	gx_hide_pointer

/*
**	program display controller registers from provided values
*/

static void program_chipset (struct kgi_display *dpy, 
			     struct kgi_mode *mode, struct chipsetregs *cr) {
	IS_CRITICAL;
	struct	chipset_private *private = KGIM_PRIVATE (dpy, chipset);
	mem_vaddr	gxbase = private->gx_control.baseptr;
	mem_vaddr	cxbase = private->gx_companion.baseptr;
	kgiu32 lockval, gcfg, tcfg, ocfg, dcfg = 0, i;

	ENTER_CRITICAL;

	/*	unlock the display controller registers		*/

	lockval = READ_GX32 (gxbase, DC_UNLOCK);
	WRITE_GX32 (gxbase, DC_UNLOCK, DC_UNLOCK_VALUE);

	/*	blank display and disable clock			*/

	gcfg = READ_GX32 (gxbase, DC_GENERAL_CFG);
	tcfg = READ_GX32 (gxbase, DC_TIMING_CFG);
	ocfg = READ_GX32 (gxbase, DC_OUTPUT_CFG);

	if (private->companion) {
		dcfg  = READ_GX32 (cxbase, CX5520_DISPLAY_CONFIG);
		dcfg &= ~(kgiu32) CX5520_DCFG_DAC_BL_EN;
		WRITE_GX32 (cxbase, CX5520_DISPLAY_CONFIG, dcfg);
	}

	tcfg &= ~(kgiu32) DC_TCFG_BLKE;
	WRITE_GX32 (gxbase, DC_TIMING_CFG, tcfg);

	tcfg &= ~(kgiu32) DC_TCFG_TGEN;
	WRITE_GX32 (gxbase, DC_TIMING_CFG, tcfg);

	/*	disable display fifo load and compression	*/
    
	gcfg &= ~(kgiu32) (DC_GCFG_DFLE | DC_GCFG_CMPE | DC_GCFG_DECE);
	WRITE_GX32 (gxbase, DC_GENERAL_CFG, gcfg);

	/*	clear dotclock multiplication			*/

	gcfg &= ~(kgiu32) (DC_GCFG_DDCK | DC_GCFG_DPCK | DC_GCFG_DFCK);
	gcfg &= ~(kgiu32) DC_GCFG_DCLK_MASK;
	WRITE_GX32 (gxbase, DC_GENERAL_CFG, gcfg);

	/*	program ramdac and dotclock			*/

	kgim_monitor_set_mode (dpy, mode);
	kgim_clock_set_mode (dpy, mode);
	kgim_ramdac_set_mode (dpy, mode);

	/*	set dotclock multiplication			*/

	gcfg |= cr->general_cfg & DC_GCFG_DCLK_MASK;
	WRITE_GX32 (gxbase, DC_GENERAL_CFG, gcfg);

	/*	write display controller regs			*/

	gcfg  = cr->general_cfg;
	gcfg &= ~(kgiu32) (DC_GCFG_DECE | DC_GCFG_CMPE);
	tcfg  = cr->timing_cfg;
	ocfg  = cr->output_cfg;

	WRITE_GX32 (gxbase, DC_FB_ST_OFFSET, cr->fb_st_offset);
	WRITE_GX32 (gxbase, DC_CB_ST_OFFSET, cr->cb_st_offset);
	WRITE_GX32 (gxbase, DC_CURS_ST_OFFSET, cr->cur_st_offset);
	WRITE_GX32 (gxbase, DC_LINE_DELTA, cr->dc_line_delta);
	WRITE_GX32 (gxbase, DC_BUF_SIZE, cr->dc_buf_size);

	while (READ_GX32 (gxbase, GP_BLIT_STATUS) & BS_BLIT_BUSY);
	WRITE_GX32 (gxbase, GP_BLIT_STATUS, cr->gp_blit_status);

	WRITE_GX32 (gxbase, DC_H_TIMING_1, cr->h_timing_1);
	WRITE_GX32 (gxbase, DC_H_TIMING_2, cr->h_timing_2);
	WRITE_GX32 (gxbase, DC_H_TIMING_3, cr->h_timing_3);
	WRITE_GX32 (gxbase, DC_FP_H_TIMING, cr->h_timing_3);
	WRITE_GX32 (gxbase, DC_V_TIMING_1, cr->v_timing_1);
	WRITE_GX32 (gxbase, DC_V_TIMING_2, cr->v_timing_2);
	WRITE_GX32 (gxbase, DC_V_TIMING_3, cr->v_timing_3);
	WRITE_GX32 (gxbase, DC_FP_V_TIMING, cr->v_timing_3);

	if (private->companion) {
		/*	Set h/v sync polarities			*/

		dcfg &= ~(kgiu32) CX5520_DCFG_CRT_HSYNC_POL;
		dcfg &= ~(kgiu32) CX5520_DCFG_CRT_VSYNC_POL;
		if (tcfg & DC_TCFG_CHSP) dcfg |= CX5520_DCFG_CRT_HSYNC_POL;
		if (tcfg & DC_TCFG_CVSP) dcfg |= CX5520_DCFG_CRT_VSYNC_POL;
		tcfg &= ~(kgiu32) (DC_TCFG_CHSP | DC_TCFG_CVSP);
	}

	WRITE_GX32 (gxbase, DC_OUTPUT_CFG, ocfg);
	WRITE_GX32 (gxbase, DC_TIMING_CFG, tcfg);
	WRITE_GX32 (gxbase, DC_GENERAL_CFG, gcfg);

	if (private->companion) {
		/*	Enable display and CRT output		*/

		dcfg |= CX5520_DCFG_DIS_EN;
		dcfg |= CX5520_DCFG_HSYNC_EN | CX5520_DCFG_VSYNC_EN;
		dcfg |= CX5520_DCFG_DAC_BL_EN | CX5520_DCFG_DAC_PWDNX;

		/*	Let graphics bypass secondary palette	*/

		dcfg |= CX5520_DCFG_GV_PAL_BYP;

		WRITE_GX32 (cxbase, CX5520_DISPLAY_CONFIG, dcfg);
	}

	/*	prepare and enable compression if requested	*/

	if (cr->general_cfg & (DC_GCFG_DECE|DC_GCFG_CMPE)) {

		/*	clear compressed line "valid" bits	*/

		for (i=0; i < 1024; i++) {
			WRITE_GX32 (gxbase, MC_DR_ADD, i);
			WRITE_GX32 (gxbase, MC_DR_ACC, 0);
		}

		/*	enable framebuffer compression		*/

		gcfg |= (DC_GCFG_DECE|DC_GCFG_CMPE);
		WRITE_GX32 (gxbase, DC_GENERAL_CFG, gcfg);
	}

	/*	write line compare and cursor registers		*/

	WRITE_GX32 (gxbase, DC_SS_LINE_COMPARE, cr->ss_line_compare);
	WRITE_GX32 (gxbase, DC_CURSOR_X, cr->dc_cursor_x);
	WRITE_GX32 (gxbase, DC_CURSOR_Y, cr->dc_cursor_y);
	WRITE_GX32 (gxbase, DC_CURSOR_COLOR, cr->dc_cursor_color);

	/*	restore status of lock register			*/

	WRITE_GX32 (gxbase, DC_UNLOCK, lockval);

	LEAVE_CRITICAL;
}
