/* $Id: mga_common.c,v 1.13 1999/05/25 21:29:17 marcus Exp $
******************************************************************************

   Matrox Mystique and Millennium common functions

   Copyright (C) 1996-1997 Jason McMullan	[jmcc@ggi-project.org]

   Permission is hereby granted, free of charge, to any person obtaining a
   copy of this software and associated documentation files (the "Software"),
   to deal in the Software without restriction, including without limitation
   the rights to use, copy, modify, merge, publish, distribute, sublicense,
   and/or sell copies of the Software, and to permit persons to whom the
   Software is furnished to do so, subject to the following conditions:

   The above copyright notice and this permission notice shall be included in
   all copies or substantial portions of the Software.

   THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
   IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
   FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL
   THE AUTHOR(S) BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
   IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
   CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

******************************************************************************

   This code was derived from the following sources of information:

   Matrox MGA-1064SG Developer Specification (Internet edition)
   Matrox Millenium mga2064.c by Jonas Borgstrm

******************************************************************************
*/

#include <kgi/kgi.h>
#include <kgi/module.h>
#include <linux/fb.h>

#include <Matrox/mga_common_regs.h>  /* Register definitions */
#include <Matrox/mga_common_mmio.h>
#include <Matrox/mga_common_text16.h>
#include <Matrox/mga_common.h>
/* Driver specific things are included by mga_common_*.h */

/*
** Rounds up virt and visible fields to the next multiple of 'round'.
** Enlarges virt to be at least of size visible.
*/
static void roundmode(kgi_modereq *gm,int roundx,int roundy)
{
	gm->visible.x = ((gm->visible.x + roundx - 1) / roundx) * roundx;
	gm->visible.y = ((gm->visible.y + roundy - 1) / roundy) * roundy;
	gm->virt.x    = ((gm->virt.x + roundx - 1) / roundx) * roundx;
	gm->virt.y    = ((gm->virt.y + roundy - 1) / roundy) * roundy;

	if (gm->visible.x > gm->virt.x)
 	        gm->virt.x = gm->visible.x;
	if (gm->visible.y > gm->virt.y)
	        gm->virt.y = gm->visible.y;
}

/*
** determines the available RAM on the card.  Primitive but functional.
*/
static unsigned long mga_countram(struct kgi_display *dpy)
{
        GET_PRIVATE;
	IS_CRITICAL;
        kgiu8 tmp, tmp3mb, tmp5mb;

	ENTER_CRITICAL;
        
	/*      Enable framebuffer
        */
	mem_out8(CRTX_EXT3, MMIO_CRTXI);
        tmp = mem_in8(MMIO_CRTXD);
        mem_out8(tmp | CRTX_MGAMODE, MMIO_CRTXD);

        /*      Write and reread test values
	**      Must take care for a wrap around.  The trick is
	**      that the mem_out8()'s overwrite the preceding ones.
	*/
        mem_out8(0x55, MGA_BASE2 + 0x500000);
        mem_out8(0x33, MGA_BASE2 + 0x300000);
	mem_out8(0x11, MGA_BASE2 + 0x100000);
        tmp5mb = mem_in8(MGA_BASE2 + 0x500000);
        tmp3mb = mem_in8(MGA_BASE2 + 0x300000);
      
        /*      Restore framebuffer state
        */
	mem_out8(CRTX_EXT3, MMIO_CRTXI);
        mem_out8(tmp, MMIO_CRTXD);

	LEAVE_CRITICAL;
	
	return 1024 * ((tmp5mb == 0x55) ? 8192 : (tmp3mb == 0x33 ? 4096 : 2048));
}

/*
** Checks for a (Mystique/Millenium (I/II)) and initializes values. Includes
** - chipinfo.ramsize
** - graphic_mmio_linear_fb
** - kgi_display->private->cp (fully initialized after)
**
** Claims memory regions.
*/
static int mga_verify(struct kgi_display *dpy, int productID)
{
	GET_PRIVATE;
	kgiu8 rev;
	kgiu8 bus;
	kgiu8 dev_fn;
	kgiu16 com;
	kgiu32 opt;
	int   ctrloff, fboff;

	MGA_CP.chipinfo = chipinfo;

	/* !!! Using Linux functions until decided which */
	/* PCI functions to use */
	if (pcibios_find_device(PCI_VENDOR_ID_MATROX, productID,
				0, /* !!! Multihead */
				&bus, &dev_fn)) {
			NOTICE("Cannot find a Matrox 0x%x\n", productID);
                return -E(CHIPSET, UNKNOWN);
	}
	
        MGA_PCI_BUS    = bus;
        MGA_PCI_DEV_FN = dev_fn;
        MGA_CHIP_ID    = productID;

	/* Enable device response and disable decoding of I/O and
	** VGA framebuffer locations.  This MUST come first, because
	** on a disabled card the PCI BIOS may choose to change
	** memory addresses! (VGA framebuffer in mga??64_main.c)
	*/
	pcibios_read_config_word(bus, dev_fn, PCI_COMMAND, &com);
	MGA_COLDINIT = (com & PCI_COMMAND_MEMORY) == 0;
#if 1
	com = (com & ~PCI_COMMAND_IO) | PCI_COMMAND_MEMORY;
#endif
	pcibios_write_config_word(bus, dev_fn, PCI_COMMAND, com);
	pcibios_read_config_dword(bus, dev_fn, MGA_PCI_OPTION, &opt);
#if 1
	opt &= ~MGA_PCI_OPTION_VGAIOEN;
#endif
#if defined(__BIG_ENDIAN)
	/* Enable byte swapping for big endian systmes for register space */
	opt |= MGA_PCI_OPTION_POWERPC; /* Not enough though */
#else
	/* Assume we are on a little endian system */
	opt &= ~MGA_PCI_OPTION_POWERPC;
#endif
	pcibios_write_config_dword(bus, dev_fn, MGA_PCI_OPTION, opt);
	/* !!! Disable VGA framebuffer (0xA0000 - 0xC0000) */
	
        pcibios_read_config_byte(bus, dev_fn, PCI_CLASS_REVISION, &rev);
        DEBUG1("%s card (rev %3d) detected.", MGA_CP.chipinfo.model, rev);
	MGA_CP.rev = rev;

        MGA_APERTURE    = mga_aperture;
        MGA_FRAMEBUFFER = mga_framebuffer;
	
	if (productID == PCI_DEVICE_ID_MATROX_MIL_2 || 
	    productID == PCI_DEVICE_ID_MATROX_MIL_2_AGP || 
	    (productID == PCI_DEVICE_ID_MATROX_MYS && rev >= 3)) {
		ctrloff = PCI_BASE_ADDRESS_1;
		fboff   = PCI_BASE_ADDRESS_0;
	} else {
		ctrloff = PCI_BASE_ADDRESS_0;
		fboff   = PCI_BASE_ADDRESS_1;
	}
	pcibios_read_config_dword(bus, dev_fn, ctrloff, &MGA_APERTURE.base);
	pcibios_read_config_dword(bus, dev_fn, fboff, &MGA_FRAMEBUFFER.base);
	MGA_APERTURE.base    &= 0xffffc000;
        MGA_FRAMEBUFFER.base &= 0xff800000;

        if (mem_check_region(&MGA_APERTURE) || 
	    mem_check_region(&MGA_FRAMEBUFFER)) {
                ERROR("  FRAMEBUFFER or APERTURE busy.");
                return -E(CHIPSET, UNKNOWN);
        }
        mem_claim_region(&MGA_APERTURE);
        mem_claim_region(&MGA_FRAMEBUFFER);

        MGA_CP.chipinfo.ramsize = mga_countram(dpy);
        DEBUG1("Linear framebuffer = 0x%x Byte memory (%p) (%08x)",
	       MGA_CP.chipinfo.ramsize, MGA_FRAMEBUFFER.baseptr, 
	       MGA_FRAMEBUFFER.base);

	MGA_CP.graphic_mmio_linear_fb        = graphic_mmio_linear_fb;
        MGA_CP.graphic_mmio_linear_fb.pstart = MGA_FRAMEBUFFER.base;
        MGA_CP.graphic_mmio_linear_fb.size   =
        MGA_CP.graphic_mmio_linear_fb.psize  = MGA_CP.chipinfo.ramsize;

	/* Initialize IRQ */
	MGA_CP.vsync_wait = NULL;
	pcibios_read_config_byte(bus, dev_fn, PCI_INTERRUPT_LINE, &MGA_IRQ);
	if (MGA_IRQ != 255) {
		DEBUG1("setting IRQ %d", MGA_IRQ);
		mem_out32(0, IEN);    /* Disable interrupts */
		mem_out32(0, ICLEAR); /* Clear interrupts */
		CRT_out8(dpy, (CRT_in8(dpy, CRT_VSYNCEND) & 
			       ~CR11_CLEAR_VSYNC_IRQ) | CR11_DISABLE_VSYNC_IRQ,
			 /* Only useful for the Mystique, but doesn't hurt the Millenium */
			 CRT_VSYNCEND); /* Clear and disable vsync int */

		if (request_irq(MGA_IRQ, mga_interrupt_handler, SA_SHIRQ,
				MGA_CP.chipinfo.model, dpy)) {
			DEBUG1("Couldn't set interrupt");
			MGA_IRQ = 255;
		}
	}
        
	/* Set interleave in the in the Millenium code */

        private->cp.ptr_and = 0x00FF;
        private->cp.ptr_xor = 0x7F00;
        private->cp.cur_and = 0x00FF;
        private->cp.cur_xor = 0x6F00;
   
	/* Reusing boot display framebuffer operations */
	mgatext_color_io.a_to_p  = text16_color_io.a_to_p;
	mgatext_color_io.p_to_a  = text16_color_io.p_to_a;
	mgatext_color_io.f_to_p  = text16_color_io.f_to_p;
	mgatext_color_io.p_to_f  = text16_color_io.p_to_f;

	if (MGA_COLDINIT) {
		return mga_cold_init(dpy);
	} else {
		return -EOK;
	}
}

/*
**      KGI module interface
*/

/*
** Common initialization.
**
** Note: The chipset special init function kgim_chipset_init() MUST
** call this function first without doing any initialization beforehand!
** Otherwise you'll run into trouble, if init fails.
*/
int mga_chipset_init(struct kgi_card *card, struct kgi_display *dpy,
		     int privsize, int productID)
{
        struct chipset_private *private;
   
        KGIM_PRIVATE(dpy, chipset) = NULL;
        private = kgi_malloc_small(privsize);
        if (private == NULL) {
                NOTICE2("Failed to allocate private data");
                return -ENOMEM;
        }
        KGIM_PRIVATE(dpy, chipset) = private;

	/* Important if init fails! */
        memset(private, 0, privsize);
	MGA_IRQ = 255;
   
        if (mga_verify(dpy, productID) != EOK) {
		mga_chipset_done(dpy);
                return -E(CHIPSET, FAILED);
        }
        card->chipset  = &MGA_CP.chipinfo;
	strcpy(card->manufact,MGA_CP.chipinfo.manufact);
	strcpy(card->model,MGA_CP.chipinfo.model);
	card->modes    = MGA_CP.chipinfo.modes;
	card->flags    = MGA_CP.chipinfo.flags;
	card->maxdots  = MGA_CP.chipinfo.maxdots;

        NOTICE("%s", chipset_name_init);
        return EOK;
}

/*
** Releases common allocated resources.
**
** Note: all resources which are to be freed from this procedure must be 
** protected against uninitialized freeing.  For example, set all unset 
** pointers to NULL, and this pointer won't be released.
**
** It requires however that private is already allocated.
**
** The purpose is to make this function callable from anywhere within the 
** chipset init function.
*/
void mga_chipset_done(struct kgi_display *dpy)
{
        GET_PRIVATE;

	if (private) {
		if (MGA_FRAMEBUFFER.baseptr) {
			mem_out32(0, IEN);    /* Disable interrupts */
			mem_out32(0, ICLEAR); /* Clear interrupts */
			if (MGA_IRQ != 255) {
				CRT_out8(dpy, (CRT_in8(dpy, CRT_VSYNCEND) & 
					       ~CR11_CLEAR_VSYNC_IRQ) | 
					 CR11_DISABLE_VSYNC_IRQ,
					 CRT_VSYNCEND);
				/* Only useful for the Mystique, but */
				/* doesn't hurt the Millenium. (does it?) */
				free_irq(MGA_IRQ, dpy);
			}

			mem_free_region(&MGA_APERTURE);
			mem_free_region(&MGA_FRAMEBUFFER);
		}
		if (MGA_COLDINIT) {
			mga_cold_shutdown(dpy);
		}
		kgi_free_small(private);
		KGIM_PRIVATE(dpy, chipset) = NULL;
	}
}

kgi_sint kgim_chipset_check_mode
(struct kgi_display *dpy, struct kgi_mode *mode,
 enum kgi_tm_cmd *cmd)
{
        GET_PRIVATE;
        int bpp=0;
        int textx, texty;

	textx = mode->request.dpp.x;
        texty = mode->request.dpp.y;
   
        DEBUG2("Checking mode... (%dx%d %d)\n", mode->request.virt.x,
                mode->request.virt.y, mode->request.graphtype);

        switch (mode->request.graphtype) {

        case KGIGT_TEXT16: break;          /* Not used in textmode */
        case KGIGT_8BIT:   bpp = 8; break;
	case KGIGT_15BIT:  /* Fall thro */
        case KGIGT_16BIT:  bpp = 16; break;
#if defined(__LITTLE_ENDIAN)
        case KGIGT_24BIT:  bpp = 24; break;
#endif
        case KGIGT_32BIT:  bpp = 32; break;         

        default:
                DEBUG1("graphtype %i not supported", mode->request.graphtype);
                return -E(CHIPSET,FAILED);
        }
   
        #define XVIRTUAL (mode->request.virt.x)
        #define YVIRTUAL (mode->request.virt.y)
        #define XVISIBLE (mode->request.visible.x)
        #define YVISIBLE (mode->request.visible.y)
   
        switch (*cmd) {

        case CMD_PROPOSE:
		if (XVISIBLE <= 0 || YVISIBLE <= 0) {
			return -EINVAL;
		}
		mode->magnify.x = mode->magnify.y = 1;
         
                if (mode->request.graphtype == KGIGT_TEXT16) {

                        if ((textx < 8) || (9 < textx)) {
                                DEBUG1("Invalid char width (%i)", textx);
                                return -EINVAL;
                        }

			roundmode(&mode->request, 1, 1);
			XVIRTUAL = (XVIRTUAL + 1) & ~1;

                        if ((((16*XVIRTUAL*YVIRTUAL) + 0xFFFF) & ~0xFFFF) >
			    MGA_CP.chipinfo.ramsize) {
                                NOTICE("Not enough text memory");
                                return -ENOMEM;
                        }
			mode->x.width = mode->request.visible.x * textx;
			mode->y.width = mode->request.visible.y * texty;

                } else {        /* mode->request.graphtype != KGIGT_TEXT16 */

			roundmode(&mode->request, 8, 1);

                        while ((XVISIBLE * mode->magnify.x) < 640) {
                                mode->magnify.x++;
                        }
         
                        while ((YVISIBLE * mode->magnify.y) < 350) {
                                mode->magnify.y++;
                        }
         
                        /*      Can't zoom more than this */
                        if (mode->magnify.y > 2) {
                                return -EINVAL;
                        }
         
                        /*      Can't do more than doublescan */
                        if (mode->magnify.x > 2) {
                                return -EINVAL;
                        }
         
			/* Adjust virtual size */
			if (adjust_virtual(mode)) {
				DEBUG1("x virtual size %d too big",
				       mode->request.virt.x);
			        return -EINVAL;
			}

			if ((((XVIRTUAL * YVIRTUAL * bpp/8) + 
			      0xFFFF) & ~0xFFFF) > MGA_CP.chipinfo.ramsize) {
				DEBUG1("Not enough memory to do "
				       "virtual %ix%i at %i bpp",
				       XVIRTUAL, YVIRTUAL, bpp);
				return -ENOMEM;
                        }
			mode->x.width = mode->request.visible.x * 
			  mode->magnify.x;
			mode->y.width = mode->request.visible.y * 
			  mode->magnify.y;
                }
      
                return EOK;
      
        case CMD_LOWER:

                if (mode->dclk < MGA_CP.chipinfo.dclk.min) {
                        DEBUG1("DCLK too low (%i)", mode->dclk);
                        return -E(CHIPSET,FAILED);
                }

                mode->dclk = (mode->dclk > MGA_CP.chipinfo.dclk.max)
                        ? MGA_CP.chipinfo.dclk.max : mode->dclk;
      
                adjust_timing(mode);
                return EOK;
      
        case CMD_RAISE:
      
                if (mode->dclk > MGA_CP.chipinfo.dclk.max) {
                        DEBUG1("DCLK too high (%i)", mode->dclk);
                        return -E(CHIPSET,FAILED);
                }

                mode->dclk = (mode->dclk < MGA_CP.chipinfo.dclk.min)
                        ? MGA_CP.chipinfo.dclk.min : mode->dclk;
      
                adjust_timing(mode);
                return EOK;
      
        case CMD_CHECK:

                if (mode->lclk.mul != 1 && mode->lclk.mul != 3) {
                        DEBUG1("Invalid lclk multiplier %d", mode->lclk.mul);
                        return -E(CHIPSET,FAILED);
                }

		switch (mode->lclk.div) {

		case 1:
		case 2:
		case 4:
		case 8:
		case 16:
		case 32:
			break;

		default:
			DEBUG1("invalid CLK to LCLK divider %i", mode->lclk.div);
			return -E(CHIPSET, UNKNOWN);
		}

#define INTERLEAVE (dpy->card->chipset->ramsize > 2 MB)
		if(MGA_CHIP_ID != PCI_DEVICE_ID_MATROX_MYS && !INTERLEAVE)
                {
			switch(mode->request.graphtype)
			{
			case KGIGT_8BIT:
	                	mode->lclk.div = 4; break;
			case KGIGT_15BIT:
			case KGIGT_16BIT:
	                	mode->lclk.div = 2; break;
#if defined(__LITTLE_ENDIAN)
			case KGIGT_24BIT:
	                	mode->lclk.div = 4; break;
#endif
			case KGIGT_32BIT:
	                	mode->lclk.div = 1; break;
			default:
				break;
			};	
		}
#undef INTERLEAVE
		if(MGA_CHIP_ID != PCI_DEVICE_ID_MATROX_MYS)
                {
		   
			mode->clk.mul = mode->lclk.mul;
			mode->clk.div = mode->lclk.div;
		}

                if (mode->request.graphtype == KGIGT_TEXT16) {
                        mode->flags |= FL_CHIP_COLORTEXT;
                        mode->fsize = XVIRTUAL * YVIRTUAL * 2;
			mode->fb_size = (mode->fsize*8 + 0xFFFF) & ~0xFFFF; /* !!! */
                        mode->mem   = &text16_color_mem;
                        mode->io    = &mgatext_color_io;

                        if ((mode->magnify.x != 1) || (mode->magnify.y != 1) ||
			    (textx < 8) || (9 < textx) || (texty < 1) ||
			    (texty > 32) || (XVIRTUAL&1) != 0 ||
			    (mode->fb_size > MGA_CP.chipinfo.ramsize) ||
			    (mode->dclk > MGA_CP.chipinfo.dclk.max) ||
			    (mode->dclk < MGA_CP.chipinfo.dclk.min)) {
                                NOTICE1("HW check failed");
                                return -E(CHIPSET,FAILED);
                        }

                } else {         

                        mode->fsize  = (XVIRTUAL * YVIRTUAL * bpp) / 8;
			mode->fb_size = (mode->fsize + 0xFFFF) & ~0xFFFF; /* !!! */
                        if ((mode->fb_size > MGA_CP.chipinfo.ramsize) ||
			    (mode->magnify.x > 2) || (mode->magnify.y > 2) ||
			    (mode->dclk < MGA_CP.chipinfo.dclk.min) ||
			    (MGA_CP.chipinfo.dclk.max < mode->dclk) ||
			    (mode->magnify.x > 2) || (mode->magnify.y > 2) ||
			    ((mode->request.graphtype != KGIGT_8BIT) &&
			     (mode->request.graphtype != KGIGT_16BIT) &&
			     (mode->request.graphtype != KGIGT_15BIT) &&
			     (mode->request.graphtype != KGIGT_24BIT) &&
			     (mode->request.graphtype != KGIGT_32BIT))) {
				DEBUG1("Sanity check failed\n");
				return -E(CHIPSET,FAILED);
                        }
                        mode->mem = NULL;
                        mode->io  = NULL;
                }
		mode->fb = (char *)MGA_FRAMEBUFFER.baseptr;
		/* We should always map the entire framebuffer. */
		mode->fb_size = MGA_CP.chipinfo.ramsize;
		MGA_CP.graphic_mmio_linear_fb.size  = mode->fb_size;
		MGA_CP.graphic_mmio_linear_fb.psize = mode->fb_size;
		mode->mmio  = &MGA_CP.graphic_mmio_linear_fb;
	   
		/* For KGI mmio accel */
		mode->fbcon_mmio_start = MGA_APERTURE.base;
		mode->fbcon_mmio_len = MGA_APERTURE.size;
	   
		switch(MGA_CHIP_ID) {
			case PCI_DEVICE_ID_MATROX_MIL:
				mode->fbcon_accel = FB_ACCEL_MATROX_MGA2064W;
				break;
		   
			case PCI_DEVICE_ID_MATROX_MIL_2:
				mode->fbcon_accel = FB_ACCEL_MATROX_MGA2164W;
				break;
		   
			case PCI_DEVICE_ID_MATROX_MIL_2_AGP:
				mode->fbcon_accel = FB_ACCEL_MATROX_MGA2164W_AGP;
				break;
		   
			case PCI_DEVICE_ID_MATROX_MYS:
				mode->fbcon_accel = FB_ACCEL_MATROX_MGA1064SG;
				break;

			default:
		                INTERNAL_ERROR;
		                return -E(CHIPSET,FAILED);
		   		break;		   
		}

#ifdef DRIVER_mga1064 /* The mga2x64 driver doesn't support pingpong buffers yet */
		mode->accel = PINGPONG;
#endif      
                #define TM(X)   mode->x.X, mode->y.X
                if ((MGA_CP.chipinfo.maxdots.x < mode->x.width) ||
		    (MGA_CP.chipinfo.maxdots.y < mode->y.width)) {
                        NOTICE1("Timing check failed");
                        DEBUG1("Maxdots are (%d,%d)", 
			       MGA_CP.chipinfo.maxdots.x,
			       MGA_CP.chipinfo.maxdots.y);
                        DEBUG1("%ix%i %ix%i %ix%i %ix%i %ix%i %ix%i\n",
                                TM(width), TM(blankstart), TM(syncstart),
                                TM(syncend), TM(blankend), TM(total));

			return -E(CHIPSET,FAILED);
                }
                #undef TM
                return EOK;
      
        default:
                INTERNAL_ERROR;
                return -E(CHIPSET,FAILED);
        }
   
        #undef XVIRTUAL
        #undef YVIRTUAL
        #undef XVISIBLE
        #undef YVISIBLE
}
